/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.duanledexianxianxian.maven.flink;

import com.alibaba.fastjson.JSONObject;
import com.duanledexianxianxian.maven.flink.model.Student;
import com.google.common.collect.Lists;
import org.apache.flink.api.common.functions.AggregateFunction;
import org.apache.flink.api.common.functions.FlatMapFunction;
import org.apache.flink.api.common.functions.MapFunction;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.functions.KeySelector;
import org.apache.flink.api.java.tuple.Tuple2;
import org.apache.flink.streaming.api.TimeCharacteristic;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.KeyedStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.util.Collector;

import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

/**
 * Skeleton for a Flink Streaming Job.
 *
 * <p>For a tutorial how to write a Flink streaming application, check the
 * tutorials and examples on the <a href="http://flink.apache.org/docs/stable/">Flink Website</a>.
 *
 * <p>To package your application into a JAR file for execution, run
 * 'mvn clean package' on the command line.
 *
 * <p>If you change the name of the main class (with the public static void main(String[] args))
 * method, change the respective entry in the POM.xml file (simply search for 'mainClass').
 *
 * @author Administrator
 */
public class StreamingJob {

    public static void main(String[] args) throws Exception {
        // 创建执行环境上下文
        final StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        //设置窗口的时间单位为process time
        env.setStreamTimeCharacteristic(TimeCharacteristic.IngestionTime);
        Properties properties = new Properties();
        properties.setProperty("bootstrap.servers", "113.105.144.9:9104");


        FlinkKafkaConsumer<String> consumer = new FlinkKafkaConsumer<>("topic_first", new SimpleStringSchema(), properties);

        //从最早开始消费
//        consumer.setStartFromEarliest();
        DataStream<String> stream = env
                .addSource(consumer);
//        stream.print();

        // map 函数  flatMap是map的特殊形式
        SingleOutputStreamOperator<Student> student = stream.map((MapFunction<String, Student>) value -> JSONObject.parseObject(value, Student.class));
//        student.print();

        // flatMap函数
//        SingleOutputStreamOperator<String> studentFlatMap = student.flatMap((FlatMapFunction<Student, String>) (value, out) -> {
//            for (String word : value.getName().split("")) {
//                out.collect(word);
//            }
//        }).returns(String.class);
//        studentFlatMap.print();


        student.keyBy(Student::getSex).window(TumblingEventTimeWindows.of(Time.seconds(20))).aggregate(new Agg()).print();



        /*
         * Here, you can start creating your execution plan for Flink.
         *
         * Start with getting some data from the environment, like
         * 	env.readTextFile(textPath);
         *
         * then, transform the resulting DataStream<String> using operations
         * like
         * 	.filter()
         * 	.flatMap()
         * 	.join()
         * 	.coGroup()
         *
         * and many more.
         * Have a look at the programming guide for the Java API:
         *
         * http://flink.apache.org/docs/latest/apis/streaming/index.html
         *
         */


        // execute program
        env.execute("Flink Streaming Java API Skeleton");
    }

    /**
     * 自定义聚合函数
     */
    public static class Agg implements AggregateFunction<Student, Tuple2<List<Student>, Long>, Tuple2<List<Student>, Long>> {
        //创建一个数据统计的容器，提供给后续操作使用。
        @Override
        public Tuple2<List<Student>, Long> createAccumulator() {
            return new Tuple2<List<Student>, Long>();
        }

        //每个元素被添加进窗口的时候调用。
        //第一个参数是添加进窗口的元素，第二个参数是统计的容器(上面创建的那个)。
        @Override
        public Tuple2<List<Student>, Long> add(Student in, Tuple2<List<Student>, Long> acc) {
            List<Student> list = acc.f0;
            Long count = acc.f1;
            if (list == null) {
                list = Lists.newArrayList();
            }
            list.add(in);
            if (count == null) {
                count = 1L;
            } else {
                count++;
            }
            return new Tuple2<>(list, count);
        }

        //窗口统计事件触发时调用来返回出统计的结果。
        @Override
        public Tuple2<List<Student>, Long> getResult(Tuple2<List<Student>, Long> acc) {
            return acc;
        }

        //只有在当窗口合并的时候调用,合并2个容器
        @Override
        public Tuple2<List<Student>, Long> merge(Tuple2<List<Student>, Long> acc1, Tuple2<List<Student>, Long> acc2) {
            List<Student> list1 = acc1.f0;
            List<Student> list2 = acc2.f0;
            Long count1 = acc1.f1;
            Long count2 = acc2.f1;
            list1.addAll(list2);
            return new Tuple2<>(list1, count1 + count2);
        }
    }
}


